"""Module for manually triggering a step function"""
# Standard
import argparse
from datetime import datetime, timezone
import json
import time
import logging
# Installed
import boto3
from botocore.exceptions import ClientError
# Local
from libera_utils.logutil import configure_task_logging
from libera_utils.aws import utils

logger = logging.getLogger(__name__)


def step_function_trigger(parsed_args: argparse.Namespace):
    """Start a stepfunction to process a certain days data
        Parameters
        ----------
        parsed_args : argparse.Namespace
            Namespace of parsed CLI arguments

        region_name : str
            string of the AWS region name

        Returns
        -------
        None
    """
    now = datetime.now(timezone.utc)
    configure_task_logging(f'processing_step_function_trigger_{now}',
                           limit_debug_loggers='libera_utils',
                           console_log_level=logging.DEBUG)

    logger.debug(f"CLI args: {parsed_args}")

    region_name = 'us-west-2'

    account_id = utils.get_aws_account_number()
    logger.debug(f"Account id is : {account_id}")

    state_machine_arn = f"arn:aws:states:{region_name}:{account_id}:stateMachine:{parsed_args.algorithm_name}"
    logger.debug(f"State machine is: {state_machine_arn}")

    step_function_client = boto3.client("stepfunctions", region_name)
    input_object = json.dumps({
        "TRIGGER_SOURCE": "Manual",
        "RETRY_COUNT": 1,
        "RETRY_HISTORY": "NA",
        "APPLICABLE_DAY": parsed_args.applicable_day})
    logger.debug(f"Input object to the state machine is : {input_object}")

    try:
        response = step_function_client.start_execution(stateMachineArn=state_machine_arn, input=input_object)
        if parsed_args.verbose:
            logger.debug("Execution Started")
        execution_response = step_function_client.describe_execution(executionArn=response['executionArn'])
        if execution_response['status'] == "RUNNING":
            if parsed_args.verbose:
                logger.debug("Waiting for execution to complete")
            if parsed_args.wait_for_finish:
                while execution_response['status'] == "RUNNING":
                    time.sleep(5)
                    execution_response = step_function_client.describe_execution(executionArn=response['executionArn'])

    except ClientError as err:
        logger.error(
            f"Couldn't start state machine {state_machine_arn}. Here's why: {err.response['Error']['Code']}: "
            f"{err.response['Error']['Message']}")
        raise

    if parsed_args.verbose:
        logger.info(f"Reached the try loop else with ARN: {response['executionArn']}")
    if execution_response['status'] == "SUCCEEDED":
        logger.info("Execution of Step Function Succeeded")

    elif execution_response['status'] == "FAILED":
        logger.info("Execution of Step Function Failed")

    else:
        logger.info(f"Function complete step function status: {execution_response['status']}")
        logger.info("See AWS console for full details on Step ")
    if parsed_args.verbose:
        logger.debug(execution_response)



"""AWS ECR Repository/Algorithm names"""
from enum import Enum


class ManifestType(Enum):
    """Enumerated legal manifest type values"""
    INPUT = 'INPUT'
    input = INPUT
    OUTPUT = 'OUTPUT'
    output = OUTPUT


class DataProductIdentifier(Enum):
    """Enumeration of data product canonical IDs used in AWS resource naming
    These IDs refer to the data products (files) themselves, NOT the processing steps (since processing steps
    may produce multiple products).

    In general these names are of the form <level>-<source>-<type>
    """
    # L0 construction record
    l0_cr = "l0-cr"

    # L0 PDS files
    l0_rad_pds = "l0-rad-pds"
    l0_cam_pds = "l0-cam-pds"
    l0_azel_pds = "l0-azel-pds"
    l0_jpss_pds = "l0-jpss-pds"

    # SPICE kernels
    spice_az_ck = "spice-az-ck"
    spice_el_ck = "spice-el-ck"
    spice_jpss_ck = "spice-jpss-ck"
    spice_jpss_spk = "spice-jpss-spk"

    # Calibration products
    cal_rad = "cal-rad"
    cal_cam = "cal-cam"

    # L1B products
    l1b_rad = "l1b-rad"
    l1b_cam = "l1b-cam"

    # L2 products
    # TODO: L2 product IDs TBD
    # l2_unf = "l2-unf"  # unfiltered radiance
    # l2_cf = "l2-cf"  # cloud fraction
    # l2_ssw_toa = "l2-ssw-toa"  # SSW TOA flux
    # l2_ssw_surf = "l2-ssw-surf"  # SSW surface flux
    # l2_fir_toa = "l2-fir-toa"  # FIR TOA flux

    # Ancillary products
    anc_adm = "anc-adm"


class ProcessingStepIdentifier(Enum):
    """Enumeration of processing step IDs used in AWS resource naming and processing orchestration

    In orchestration code, these are used as "NodeID" values to identify processing steps:
        The processing_step_node_id values used in libera_cdk deployment stackbuilder module
        and the node names in processing_system_dag.json must match these.
    They must also be passed to the ecr_upload module called by some libera_cdk integration tests.
    """
    l2cf = 'l2-cloud-fraction'
    l2_stf = 'l2-ssw-toa'
    adms = 'libera-adms'
    l2_surface_flux = 'l2-ssw-surface-flux'
    l2_firf = 'l2-far-ir-toa-flux'
    unfilt = 'l1c-unfiltered'
    spice_azel = 'spice-azel'
    spice_jpss = 'spice-jpss'
    l1b_rad = 'l1b-rad'
    l1b_cam = 'l1b-cam'
    l0_jpss_pds = 'l0-jpss'
    l0_azel_pds = 'l0-azel'
    l0_rad_pds = 'l0-rad'
    l0_cam_pds = 'l0-cam'
    l0_cr = 'l0-cr'


class CkObject(Enum):
    """Enum of valid CK objects"""
    JPSS = "JPSS"
    AZROT = "AZROT"
    ELSCAN = "ELSCAN"

    @property
    def data_product_id(self):
        """DataProductIdentifier for CKs associated with this CK object"""
        _product_id_map = {
            CkObject.JPSS: DataProductIdentifier.spice_jpss_ck,
            CkObject.AZROT: DataProductIdentifier.spice_az_ck,
            CkObject.ELSCAN: DataProductIdentifier.spice_el_ck
        }
        return _product_id_map[self]

    @property
    def processing_step_id(self):
        """ProcessingStepIdentifier for the processing step that produces CKs for this CK object"""
        _processing_step_id_map = {
            CkObject.JPSS: ProcessingStepIdentifier.spice_jpss,
            CkObject.AZROT: ProcessingStepIdentifier.spice_azel,
            CkObject.ELSCAN: ProcessingStepIdentifier.spice_azel
        }
        return _processing_step_id_map[self]


class SpkObject(Enum):
    """Enum of valid SPK objects"""
    JPSS = "JPSS"

    @property
    def data_product_id(self):
        """DataProductIdentifier for SPKs associated with this SPK object"""
        # Only one data product for SPKs
        return DataProductIdentifier.spice_jpss_spk

    @property
    def processing_step_id(self):
        """ProcessingStepIdentifier for the processing step that produces SPKs for this SPK object"""
        # Only one processing step that produces an SPK
        return ProcessingStepIdentifier.spice_jpss


class DataLevel(Enum):
    """Data product level"""
    L0 = "L0"
    SPICE = "SPICE"
    CAL = "CAL"
    L1B = 'L1B'
    L2 = 'L2'



"""Module for uploading docker images to the ECR"""
# Standard
import argparse
from datetime import datetime, timezone
import subprocess  # nosec
import logging
import json
# Installed
import docker
# Local
from libera_utils.logutil import configure_task_logging
from libera_utils.aws import constants, utils

logger = logging.getLogger(__name__)


def login_to_ecr(account_id, region_name):
    """Login to the AWS ECR using commands
    Parameters
    ----------
    account_id : int
        Users AWS account ID

    region_name : string
        String of the region that the users AWS account is in

    Returns
    -------
    result : CompletedProcess
        subproccess object that holds the details of the completed CLI command
    """
    ecr_path = f"{account_id}.dkr.ecr.{region_name}.amazonaws.com"
    logger.debug(f'ECR path is {ecr_path}')

    # Login to ECR using subprocess
    ecr_login_command = f"aws ecr get-login-password --region {region_name} | docker login --username AWS " \
                        f"--password-stdin {ecr_path}"
    result = subprocess.run(ecr_login_command,  # nosec
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            text=True, check=False)
    return result


def upload_image_to_ecr(parsed_args: argparse.Namespace):
    """Upload docker image to the correct ECR repository

    Parameters
    ----------
    parsed_args : argparse.Namespace
        Namespace of parsed CLI arguments

    Returns
    -------
    None
    """
    now = datetime.now(timezone.utc)
    configure_task_logging(f'ecr_upload_{now}',
                           limit_debug_loggers='libera_utils',
                           console_log_level=logging.DEBUG)
    # if parsed_args.verbose else None
    docker_client = docker.from_env()
    logger.debug(f"CLI args: {parsed_args}")
    image_name = parsed_args.image_name
    if parsed_args.image_tag:
        image_tag = parsed_args.image_tag
    else:
        docker_image = docker_client.images.get(image_name)
        image_tag = docker_image.tags

    region_name = "us-west-2"
    logger.debug(f'Region set to {region_name}')

    account_id = utils.get_aws_account_number()
    logger.debug(f'Account ID is {account_id}')

    algorithm_name = constants.ProcessingStepIdentifier(parsed_args.algorithm_name).value
    ecr_name = f"{algorithm_name}-docker-repo"
    logger.debug(f'Algorithm name is {ecr_name}')

    # ECR path
    ecr_path = f"{account_id}.dkr.ecr.{region_name}.amazonaws.com"
    logger.debug(f'ECR path is {ecr_path}')

    # Login to ECR using subprocess
    result = login_to_ecr(account_id=account_id, region_name=region_name)
    if result.returncode == 0:
        logger.info(f"Docker Login successful. STDOUT: {result.stdout}")
    else:
        logger.error(f"STDERR: {result.stderr}")
        raise RuntimeError(f"ECR login command: {result.stderr} failed.")

    # # Tag the latest libera_utils image with the ECR repo name
    docker_client.images.get(f"{image_name}:{image_tag}").tag(f"{ecr_path}/{ecr_name}")
    logger.info("Pushing image to ECR.")
    resp = docker_client.images.push(f"{ecr_path}/{ecr_name}", stream=True, decode=True)
    error_message = []
    for line in resp:
        message = json.loads(json.dumps([line]))[0]
        if parsed_args.verbose:
            logger.debug(message)

        if "error" in message:
            error_message.append(message)

    if error_message:
        logger.error(f"Errors were encountered during image push. Error was: \n{error_message}")
        return
    logger.info("Image pushed to ECR successfully.")



"""Module for the Libera SDC utilities CLI
"""
# Standard
import argparse
# Local
from libera_utils import kernel_maker
from libera_utils.aws import ecr_upload, constants
from libera_utils.aws import processing_step_function_trigger as psfn
from libera_utils.version import version as libera_utils_version


def main(cli_args: list = None):
    """Main CLI entrypoint that runs the function inferred from the specified subcommand"""
    args = parse_cli_args(cli_args)
    args.func(args)


def print_version_info(*args):
    """Print CLI version information"""
    print(f"Libera SDC utilities CLI\n\tVersion {libera_utils_version()}"
          f"\n\tCopyright 2023 University of Colorado\n\tReleased under BSD3 license")


def parse_cli_args(cli_args: list):
    """Parse CLI arguments

    Parameters
    ----------
    cli_args : list
        List of CLI arguments to parse

    Returns
    -------
    : argparse.Namespace
        Parsed arguments in a Namespace object
    """
    parser = argparse.ArgumentParser(prog="libera-utils", description="Libera SDC utilities CLI")
    parser.add_argument("--version",
                        action='store_const', dest='func', const=print_version_info,
                        help="print current version of the CLI")

    subparsers = parser.add_subparsers(description="sub-commands for libera-utils CLI")

    # make-kernel
    make_kernel_parser = subparsers.add_parser('make-kernel',
                                               help='generate SPICE kernel from telemetry data')

    make_kernel_subparsers = make_kernel_parser.add_subparsers(description="sub-commands for make-kernel sub-command")

    # TODO: the interfaces to these spice kernel makers need to be changed to accept a manifest file path, which
    #   points to the PDS files from which to generate the kernels.
    # make-kernel jpss-spk
    jpss_spk_parser = make_kernel_subparsers.add_parser('jpss-spk', help="generate JPSS SPK kernel from telemetry")
    jpss_spk_parser.set_defaults(func=kernel_maker.make_jpss_spk)
    jpss_spk_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                 help="paths to L0 packet files")
    jpss_spk_parser.add_argument('--outdir', '-o', type=str,
                                 required=True,
                                 help="output directory for generated SPK")
    jpss_spk_parser.add_argument('--overwrite', action='store_true',
                                 help="force overwriting an existing kernel if it exists")
    jpss_spk_parser.add_argument('-v', '--verbose', action='store_true',
                                 help="set DEBUG level logging output")

    # make-kernel jpss-ck
    jpss_ck_parser = make_kernel_subparsers.add_parser('jpss-ck', help="generate JPSS CK kernel from telemetry")
    jpss_ck_parser.set_defaults(func=kernel_maker.make_jpss_ck)
    jpss_ck_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                help="paths to L0 packet files")
    jpss_ck_parser.add_argument('--outdir', '-o', type=str,
                                required=True,
                                help="output directory for generated CK")
    jpss_ck_parser.add_argument('--overwrite', action='store_true',
                                help="force overwriting an existing kernel if it exists")
    jpss_ck_parser.add_argument('-v', '--verbose', action='store_true',
                                help="set DEBUG level logging output")

    # make-kernel azel-ck
    azel_ck_parser = make_kernel_subparsers.add_parser('azel-ck',
                                                       help="generate Libera Az-El CK kernel from telemetry")
    azel_ck_parser.set_defaults(func=kernel_maker.make_azel_ck)
    azel_ck_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                help="paths to L0 packet files")
    azel_ck_parser.add_argument('--azimuth', action='store_true',
                                help="generate ck for Azimuth")
    azel_ck_parser.add_argument('--elevation', action='store_true',
                                help="generate ck for Elevation")
    azel_ck_parser.add_argument('--outdir', '-o', type=str, required=True,
                                help="output directory for generated CK")
    azel_ck_parser.add_argument('--overwrite', action='store_true',
                                help="force overwriting an existing kernel if it exists")
    azel_ck_parser.add_argument('--csv', action='store_true',
                                help="the provided Az and El packet_data_filepaths are ASCII csv files instead of "
                                     "binary CCSDS")
    azel_ck_parser.add_argument('-v', '--verbose', action='store_true',
                                help="set DEBUG level logging output (otherwise set by LIBSDP_STREAM_LOG_LEVEL)")

    algorithm_names = [name.value for name in constants.ProcessingStepIdentifier]
    ecr_upload_parser = subparsers.add_parser('ecr-upload', help="Upload docker image to matching ECR repository")
    ecr_upload_parser.set_defaults(func=ecr_upload.upload_image_to_ecr)
    ecr_upload_parser.add_argument('image_name', type=str, help="Image name of image to upload (image-name:image-tag)")
    ecr_upload_parser.add_argument('image_tag', type=str, help="Image tag of image to upload (image-name:image-tag)")
    ecr_upload_parser.add_argument('algorithm_name', type=str,
                                   help=f"Algorithm name that matches an ECR repo name, "
                                        f"inputs to names:\n {algorithm_names}")
    ecr_upload_parser.add_argument('-v', '--verbose', action='store_true',
                                   help="Prints out the result of the ecr push")

    sfn_trigger_parser = subparsers.add_parser('step-function-trigger',
                                               help="Manually trigger a specific step function")
    sfn_trigger_parser.set_defaults(func=psfn.step_function_trigger)
    sfn_trigger_parser.add_argument('algorithm_name', type=str, help="Algorithm name you want to run")
    sfn_trigger_parser.add_argument('applicable_day', type=str,
                                    help="Day of data you want to rerun. Format of date: YYYY-MM-DD")
    sfn_trigger_parser.add_argument('-w', '--wait_for_finish', action='store_true',
                                    help="Block command line until step function completes (may be a long time)")
    sfn_trigger_parser.add_argument('-v', '--verbose', action='store_true',
                                    help="Prints out the result of the step_function_trigger run")

    parsed_args = parser.parse_args(cli_args)
    return parsed_args



"""Module for file naming utilities"""
# Standard
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from enum import Enum
from importlib import metadata
import re
from types import SimpleNamespace
from typing import Union, Optional, Any
from pathlib import Path
# Installed
from cloudpathlib import AnyPath, S3Path
import ulid
# Local
from libera_utils.aws.constants import (
    ManifestType,
    ProcessingStepIdentifier,
    DataProductIdentifier,
    CkObject,
    SpkObject
)
from libera_utils.time import PRINTABLE_TS_FORMAT, NUMERIC_DOY_TS_FORMAT

REVISION_TS_FORMAT = f"R{NUMERIC_DOY_TS_FORMAT}"  # Just adds an r in front

SPK_REGEX = re.compile(r"^LIBERA_(?P<spk_object>JPSS)"
                       r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                       r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                       r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                       r"_(?P<revision>R[0-9]{11})"
                       r"\.bsp$")

CK_REGEX = re.compile(r"^LIBERA_(?P<ck_object>JPSS|AZROT|ELSCAN)"
                      r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                      r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                      r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                      r"_(?P<revision>R[0-9]{11})"
                      r"\.bc$")

# L0 filename format determined by EDOS Production Data Set and Construction Record filenaming conventions
LIBERA_L0_REGEX = re.compile(r"^(?P<id_char>[PX])"
                             r"(?P<scid>[0-9]{3})"
                             r"(?P<first_apid>[0-9]{4})"
                             # In some cases at least, the last character of the fill field specifies a time (T)
                             # or session (S) based product. e.g. VIIRSSCIENCEAT
                             r"(?P<fill>.{14})"
                             r"(?P<created_time>[0-9]{11})"
                             r"(?P<numeric_id>[0-9])"
                             r"(?P<file_number>[0-9]{2})"
                             r".(?P<extension>PDR|PDS)"
                             r"(?P<signal>.XFR)?$")

LIBERA_DATA_PRODUCT_REGEX = re.compile(r"^LIBERA_(?P<data_level>L1B|L2)"
                                       r"_(?P<product_name>[^_]*)"
                                       r"_(?P<version>V[0-9]*-[0-9]*-[0-9]*(RC[0-9])?)"
                                       r"_(?P<utc_start>[0-9]{8}T[0-9]{6})"
                                       r"_(?P<utc_end>[0-9]{8}T[0-9]{6})"
                                       r"_(?P<revision>R[0-9]{11})"
                                       r"\.(?P<extension>nc|h5)$")

MANIFEST_FILE_REGEX = re.compile(r"^LIBERA"
                                 r"_(?P<manifest_type>INPUT|OUTPUT)"
                                 r"_MANIFEST"
                                 r"_(?P<ulid_code>[0-9A-HJ-NP-TV-Z]{26})"
                                 r"\.json")


class ProductName(Enum):
    """Enum of valid product names as used in filenames, defined and sourced from the LASP-ASDC ICD"""
    RAD = "RAD"
    CAM = "CAM"
    # TODO: Add the additional product names from the ICD

    @property
    def processing_step_id(self):
        """ProcessingStepIdentifier for this product name"""
        _product_name_to_processing_step_id = {
            "RAD": ProcessingStepIdentifier.l1b_rad,
            "CAM": ProcessingStepIdentifier.l1b_cam
        }
        return _product_name_to_processing_step_id[self.value]

    @property
    def data_product_id(self):
        """DataProductIdentifier for this product name"""
        _product_name_to_data_product_id = {
            "RAD": DataProductIdentifier.l1b_rad,
            "CAM": DataProductIdentifier.l1b_cam
        }
        return _product_name_to_data_product_id[self.value]


class AnyFilename:
    """Polymorphic class for creating a Filename object"""

    def __new__(cls, *args, **kwargs) -> 'AbstractValidFilename':
        for CandidateClass in (L0Filename, AttitudeKernelFilename,
                               EphemerisKernelFilename, LiberaDataProductFilename,
                               ManifestFilename):
            try:
                filename = CandidateClass(*args, **kwargs)
                return filename
            except ValueError:
                continue

        raise ValueError(f"Unable to create a valid filename from {args}. "
                         "Are you sure this is a valid Libera file name?")


class AbstractValidFilename(ABC):
    """Composition of a CloudPath/Path instance with some methods to perform
    regex validation on filenames
    """
    _regex: re.Pattern
    _fmt: str
    _required_parts: tuple

    def __init__(self, *args, **kwargs):
        self.path = AnyPath(*args, **kwargs)

    def __str__(self):
        return str(self.path)

    def __eq__(self, other):
        if self.path == other.path and self.filename_parts == other.filename_parts:
            return True
        return False

    @property
    def path(self) -> Union[Path, S3Path]:
        """Property containing the file path"""
        return self._path

    @path.setter
    def path(self, new_path: Union[str, Path, S3Path]):
        if isinstance(new_path, str):
            new_path = AnyPath(new_path)
        self.regex_match(new_path)  # validates against regex pattern
        self._path: Union[Path, S3Path] = AnyPath(new_path)

    @property
    def filename_parts(self):
        """Property that contains a namespace of filename parts"""
        return self._parse_filename_parts()

    @property
    @abstractmethod
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def archive_prefix(self):
        """Property that contains the generated prefix used for archiving, when applicable"""
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def from_filename_parts(cls,
                            *args: Any,  # Required filename parts as defined by child class
                            basepath: Union[str, Path, S3Path] = None,
                            **kwargs: Any  # Optional filename parts or options as defined by child class
                            ):
        """Abstract method that must be implemented to provide hinting for required parts"""
        raise NotImplementedError()

    @classmethod
    def _from_filename_parts(cls,
                             *,  # No positional arguments
                             basepath: Union[str, Path, S3Path] = None,
                             **parts: Any):
        """Create instance from filename parts.

        The part kwarg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : Union[str, Path, S3Path], Optional
            Allows prepending a basepath or prefix.
        parts : Any
            Passed directly to _format_filename_parts. This is a dict of variable kwargs that will differ in each
            filename class based on the required parts for that particular filename type.

        Returns
        -------
        : AbstractValidFilename
        """
        filename = cls._format_filename_parts(**parts)
        if basepath is not None:
            return cls(AnyPath(basepath) / filename)
        return cls(filename)

    @classmethod
    @abstractmethod
    def _format_filename_parts(cls, **parts):
        """Format parts into a filename

        Note: When this is implemented by concrete classes, **parts becomes a set of explicitly named arguments
        """
        raise NotImplementedError()

    @abstractmethod
    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        # Do stuff to parse the elements of d into a SimpleNamespace
        raise NotImplementedError()

    @staticmethod
    def _calculate_applicable_time(start: datetime, end: datetime):
        """Based on the start time and end time of a file, returns the applicable time (date)

        Parameters
        ----------
        start : datetime.datetime
            Start of the applicable time range
        end : datetime.datetime
            End of the applicable time range

        Returns
        -------
        : datetime.date
            The date of the mean time between start and end
        """
        # In all production processing cases, utc_start and utc_end should be midnight on consecutive days
        # The applicable date is considered to be the mean between the two, ignoring hours, minutes, seconds
        t_0 = start
        t_1 = end
        t_mean = t_0 + 0.5 * (t_1 - t_0)
        applicable_date = datetime.date(t_mean)
        return applicable_date

    def regex_match(self, path: Union[str, Path, S3Path]):
        """Parse and validate a given path against class-attribute defined regex

        Returns
        -------
        : dict
            Match group dict of filename parts
        """
        # AnyPath is polymorphic but self.path will always be a CloudPath or Path object with a name attribute.
        match = self._regex.match(path.name)  # pylint: disable=no-member
        if not match:
            raise ValueError(f"Proposed path {path} failed validation against regex pattern {self._regex}")
        return match.groupdict()

    def generate_prefixed_path(self, parent_path: Union[str, Path, S3Path]) -> Union[Path, S3Path]:
        """Generates an absolute path of the form {parent_path}/{prefix_structure}/{file_basename}
        The parent_path can be an S3 bucket or an absolute local filepath (must start with /)

        Parameters
        ----------
        parent_path : Union[str, Path, S3Path]
            Absolute path to the parent directory or S3 bucket prefix. The generated path prefix is appended to the
            parent path and followed by the file basename.

        Returns
        -------
        : pathlib.Path or cloudpathlib.s3.s3path.S3Path
        """
        if isinstance(parent_path, str):
            parent_path = AnyPath(parent_path)

        if not parent_path.is_absolute():
            raise ValueError(f"Detected relative parent_path {parent_path} passed to generate_prefixed_path. "
                             "The parent_path must be an absolute path. e.g. s3://my-bucket or /starts/with/root.")

        return parent_path / self.archive_prefix / self.path.name


class L0Filename(AbstractValidFilename):
    """Filename validation class for L0 files from EDOS."""

    _regex = LIBERA_L0_REGEX
    _fmt = "{id_char}{scid:03}{first_apid:04}{fill:A<14}{created_time}{numeric_id}{file_number:02}.{extension}{signal}"
    _apid_to_l0_ingest_processing_step_id = {
        11: ProcessingStepIdentifier.l0_jpss_pds,
        1036: ProcessingStepIdentifier.l0_rad_pds,
        1048: ProcessingStepIdentifier.l0_azel_pds,
        9999: ProcessingStepIdentifier.l0_cam_pds
    }
    _apid_to_data_product_id = {
        11: DataProductIdentifier.l0_jpss_pds,
        1036: DataProductIdentifier.l0_rad_pds,
        1048: DataProductIdentifier.l0_azel_pds,
        9999: DataProductIdentifier.l0_cam_pds
    }

    @property
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        if self.filename_parts.file_number == 0:
            return ProcessingStepIdentifier.l0_cr
        return self._apid_to_l0_ingest_processing_step_id[self.filename_parts.first_apid]

    @property
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        if self.filename_parts.file_number == 0:
            return DataProductIdentifier.l0_cr
        return self._apid_to_data_product_id[self.filename_parts.first_apid]

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for L0 archiving"""
        # Generate prefix structure
        l0_file_type = "CR" if self.filename_parts.file_number == 0 else "PDS"  # CR is always PDS file_number 0
        apid = self.filename_parts.first_apid

        # 2023-07-14: This prefix might become too large over the course of the Libera mission
        return f"{l0_file_type}/{apid:0>4}"

    @classmethod
    def from_filename_parts(cls,  # noqa pylint: disable=arguments-differ
                            id_char: str,
                            scid: int,
                            first_apid: int,
                            fill: str,
                            created_time: datetime,
                            numeric_id: int,
                            file_number: int,
                            extension: str,
                            signal: Optional[str] = None,
                            basepath: Optional[Union[str, Path, S3Path]] = None):
        """Create instance from filename parts

        This method exists primarily to expose typehinting to the user for use with the generic _from_filename_parts.
        The part names are named according to the regex for the file type.

        Parameters
        ----------
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime.datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : Optional[str]
            Optional signal suffix. Always '.XFR'
        basepath : Optional[Union[str, Path, S3Path]]
            Allows prepending a basepath or prefix.

        Returns
        -------
        : L0Filename
        """
        return cls._from_filename_parts(basepath=basepath,
                                        id_char=id_char,
                                        scid=scid,
                                        first_apid=first_apid,
                                        fill=fill,
                                        created_time=created_time,
                                        numeric_id=numeric_id,
                                        file_number=file_number,
                                        extension=extension,
                                        signal=signal)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               id_char: str,
                               scid: int,
                               first_apid: int,
                               fill: str,
                               created_time: datetime,
                               numeric_id: int,
                               file_number: int,
                               extension: str,
                               signal: Optional[str] = None):
        """Construct a path from filename parts

        Parameters
        ----------
        id_char : str
            Either P (for PDS files, Construction Records) or X (for Delivery Records)
        scid : int
            Spacecraft ID
        first_apid : int
            First APID in the file
        fill : str
            Custom string up to 14 characters long
        created_time : datetime.datetime
            Creation time of the file
        numeric_id : int
            Data set ID, 0-9, one digit
        file_number : str
            File number within the data set. Construction records are always file number zero.
        extension : str
            File name extension. Either PDR or PDS
        signal : Optional[str], Optional
            Optional signal suffix. Always '.XFR'

        Returns
        -------
        : str
            Formatted filename
        """
        signal = signal if signal else ""

        return cls._fmt.format(id_char=id_char,
                               scid=scid,
                               first_apid=first_apid,
                               fill=fill,
                               created_time=created_time.strftime(NUMERIC_DOY_TS_FORMAT),
                               numeric_id=numeric_id,
                               file_number=file_number,
                               extension=extension,
                               signal=signal)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['scid'] = int(d['scid'])
        d['first_apid'] = int(d['first_apid'])
        d['numeric_id'] = int(d['numeric_id'])
        d['file_number'] = int(d['file_number'])
        d['created_time'] = datetime.strptime(d['created_time'], NUMERIC_DOY_TS_FORMAT)
        return SimpleNamespace(**d)


class LiberaDataProductFilename(AbstractValidFilename):
    """Filename validation class for L1B and L2 science products"""

    _regex = LIBERA_DATA_PRODUCT_REGEX
    _fmt = "LIBERA_{data_level}_{product_name}_{version}_{utc_start}_{utc_end}_{revision}.{extension}"

    @property
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        return ProductName[self.filename_parts.product_name].processing_step_id

    @property
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        return ProductName[self.filename_parts.product_name].data_product_id

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for L1B and L2 archiving"""
        # Generate prefix structure
        # <product_type>/<year>/<month>/<day>
        product_name = self.filename_parts.product_name

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{product_name}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # noqa pylint: disable=arguments-differ
                            data_level: str,
                            product_name: str,
                            version: str,
                            utc_start: datetime,
                            utc_end: datetime,
                            revision: datetime,
                            extension: str = 'nc',
                            basepath: Optional[Union[str, Path, S3Path]] = None):
        """Create instance from filename parts. All keyword arguments other than basepath are required!

        This method exists primarily to expose typehinting to the user for use with the generic _from_filename_parts.
        The part names are named according to the regex for the file type.

        Parameters
        ----------
        data_level : str
            L1B or L2 identifying the level of the data product
        product_name : str
            Product type. e.g. cloud-fraction for L2 or cam for L1B. May contain anything except for underscores.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime.datetime
            First timestamp in the SPK
        utc_end : datetime.datetime
            Last timestamp in the SPK
        revision: datetime.datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)
        basepath : Optional[Union[str, Path, S3Path]]
            Allows prepending a basepath or prefix.

        Returns
        -------
        : LiberaDataProductFilename
        """
        return cls._from_filename_parts(basepath=basepath,
                                        data_level=data_level,
                                        product_name=product_name,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision,
                                        extension=extension)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               data_level: str,
                               product_name: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime,
                               extension: str):
        """Construct a path from filename parts

        Parameters
        ----------
        data_level : str
            L1B or L2
        product_name : str
            Libera instrument, cam or rad for L1B and cloud-fraction etc. for L2. May contain anything except
            for underscores.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime.datetime
            First timestamp in the SPK
        utc_end : datetime.datetime
            Last timestamp in the SPK
        revision: datetime.datetime
            Time when the file was created.
        extension : str
            File extension (.nc or .h5)

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(data_level=data_level.upper(),
                               product_name=product_name.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT),
                               extension=extension)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class ManifestFilename(AbstractValidFilename):
    """Class for naming manifest files"""

    _regex = MANIFEST_FILE_REGEX
    _fmt = "LIBERA_{manifest_type}_MANIFEST_{ulid_code}.json"

    @property
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        return None  # There is no processing step that produces manifest files

    @property
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        return None  # There is no data product ID for manifest files

    @property
    def archive_prefix(self):
        """Manifests are not archived like data products, but for convenience and ease of debugging they will be kept
        in the dropbox bucket by input/output and day they were made. This is used by the step function clean up
        function in the CDK.
        # Generate prefix structure
        # <manifest_type>/<year>/<month>/<day>
        """
        manifest_type = self.filename_parts.manifest_type.value

        applicable_date = self.filename_parts.ulid_code.datetime

        return f"{manifest_type}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # noqa pylint: disable=arguments-differ
                            manifest_type: ManifestType,
                            ulid_code: ulid.ULID,
                            basepath: Union[str, Path, S3Path] = None):
        """Create instance from filename parts.

        This method exists primarily to expose typehinting to the user for use with the generic _from_filename_parts.
        The part names are named according to the regex for the file type.

        Parameters
        ----------
        manifest_type : ManifestType
            Input or output
        ulid_code : ulid.ULID
            ULID code for use in filename parts
        basepath : Optional[Union[str, Path, S3Path]]
            Allows prepending a basepath or prefix.

        Returns
        -------
        : ManifestFilename
        """
        return cls._from_filename_parts(basepath=basepath,
                                        manifest_type=manifest_type,
                                        ulid_code=ulid_code)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               manifest_type: ManifestType,
                               ulid_code: ulid.ULID):
        """Construct a path from filename parts

        Parameters
        ----------
        manifest_type : ManifestType
            Input or output
        ulid_code : ulid.ULID
            ULID code for use in filename parts

        Returns
        -------
        : str
            Formatted filename
        """
        return cls._fmt.format(manifest_type=manifest_type.value.upper(),
                               ulid_code=ulid_code)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['manifest_type'] = ManifestType(d['manifest_type'].upper())
        d['ulid_code'] = ulid.ULID.from_str(d['ulid_code'])
        return SimpleNamespace(**d)


class EphemerisKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = SPK_REGEX
    _fmt = "LIBERA_{spk_object}_{version}_{utc_start}_{utc_end}_{revision}.bsp"

    @property
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        return SpkObject[self.filename_parts.spk_object].processing_step_id

    @property
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        return SpkObject[self.filename_parts.spk_object].data_product_id

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for SPICE archiving"""
        # Generate prefix structure
        # <type>/<year>/<month>/<day>
        spk_object = self.filename_parts.spk_object

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{spk_object}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # noqa pylint: disable=arguments-differ
                            spk_object: str,
                            version: str,
                            utc_start: datetime,
                            utc_end: datetime,
                            revision: datetime,
                            basepath: Optional[Union[str, Path, S3Path]] = None):
        """Create instance from filename parts.

        This method exists primarily to expose typehinting to the user for use with the generic _from_filename_parts.
        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        spk_object : str
            Name of object whose attitude is represented in this SPK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime.datetime
            Start time of data.
        utc_end : datetime.datetime
            End time of data.
        revision: datetime.datetime
            When the file was last revised.
        basepath : Optional[Union[str, Path, S3Path]]
            Allows prepending a basepath or prefix.

        Returns
        -------
        : EphemerisKernelFilename
        """
        return cls._from_filename_parts(basepath=basepath,
                                        spk_object=spk_object,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               spk_object: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime):
        """Format filename parts as a string

        Parameters
        ----------
        spk_object : str
            Name of object whose ephemeris is represented in this SPK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime.datetime
            Start time of data.
        utc_end : datetime.datetime
            End time of data.
        revision: datetime.datetime
            Time when the file was last revised

        Returns
        -------
        : str
        """
        return cls._fmt.format(spk_object=spk_object.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


class AttitudeKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""

    _regex = CK_REGEX
    _fmt = "LIBERA_{ck_object}_{version}_{utc_start}_{utc_end}_{revision}.bc"

    @property
    def processing_step_id(self):
        """Property that contains the ProcessingStepIdentifier that generates this file"""
        return CkObject[self.filename_parts.ck_object].processing_step_id

    @property
    def data_product_id(self):
        """Property that contains the DataProductIdentifier for this file type"""
        return CkObject[self.filename_parts.ck_object].data_product_id

    @property
    def archive_prefix(self):
        """Property that contains the generated prefix for SPICE archiving"""
        # Generate prefix structure
        # <type>/<year>/<month>/<day>
        ck_object = self.filename_parts.ck_object

        applicable_date = self._calculate_applicable_time(self.filename_parts.utc_start, self.filename_parts.utc_end)

        return f"{ck_object}/{applicable_date.year:0>4}/{applicable_date.month:0>2}/{applicable_date.day:0>2}"

    @classmethod
    def from_filename_parts(cls,  # noqa pylint: disable=arguments-differ
                            ck_object: str,
                            version: str,
                            utc_start: datetime,
                            utc_end: datetime,
                            revision: datetime,
                            basepath: Optional[Union[str, Path, S3Path]] = None):
        """Create instance from filename parts.

        This method exists primarily to expose typehinting to the user for use with the generic _from_filename_parts.
        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        ck_object : str
            Name of object whose attitude is represented in this CK.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        utc_start : datetime.datetime
            Start time of data.
        utc_end : datetime.datetime
            End time of data.
        revision: datetime.datetime
            When the file was last revised.
        basepath : Optional[Union[str, Path, S3Path]]
            Allows prepending a basepath or prefix.

        Returns
        -------
        : AttitudeKernelFilename
        """
        return cls._from_filename_parts(basepath=basepath,
                                        ck_object=ck_object,
                                        version=version,
                                        utc_start=utc_start,
                                        utc_end=utc_end,
                                        revision=revision)

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               ck_object: str,
                               version: str,
                               utc_start: datetime,
                               utc_end: datetime,
                               revision: datetime):
        """Format filename parts as a string

        Parameters
        ----------
        ck_object : str
            Name of object whose attitude is represented in this CK.
        utc_start : datetime.datetime
            Start time of data.
        utc_end : datetime.datetime
            End time of data.
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision: datetime.datetime
            When the file was last revised.

        Returns
        -------
        : str
        """
        return cls._fmt.format(ck_object=ck_object.upper(),
                               version=version.upper(),
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               revision=revision.strftime(REVISION_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : types.SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        d['revision'] = datetime.strptime(d['revision'], REVISION_TS_FORMAT)
        return SimpleNamespace(**d)


def get_current_revision_str():
    """Get the current `r%y%j%H%M%S` string for filename revisions.

    Returns
    -------
    : str
        Current (now) revision string.
    """
    return datetime.now(timezone.utc).strftime(REVISION_TS_FORMAT)


def format_semantic_version(semantic_version: str):
    """Formats a semantic version string X.Y.Z into a filename-compatible string like VX-Y-Z, for X = major version,
    Y = minor version, Z = patch.

    Result is uppercase.
    Release candidate suffixes are allowed as no strict checking is done on the contents of X, Y, or Z.
    e.g. 1.2.3rc1 becomes V1-2-3RC1

    Parameters
    ----------
    semantic_version : str
        String matching X.Y.Z where X, Y and Z are integers of any length

    Returns
    -------
    : str
    """
    major, minor, patch = semantic_version.split('.')
    return f"V{major}-{minor}-{patch}".upper()


def get_current_version_str(package_name: str):
    """Retrieve the current version of a (algorithm) package and format it for inclusion in a filename

    Parameters
    ----------
    package_name : str
        Package for which to retrieve a version string. This should be your algorithm package and it must use a
        semantic versioning scheme, configured in project metadata.

    Returns
    -------
    : str
        Version string in format vM1m2p3
    """
    semver = metadata.version(package_name)
    return format_semantic_version(semver)



"""Module for database utilities"""
# Standard
import logging
# Installed
from datetime import datetime, timezone
from boto3 import resource as boto3_resource

logger = logging.getLogger(__name__)


def get_dynamodb_table(dynamo_table_name: str):
    """Get the DynamoDB table"""
    logger.debug(f"Getting DynamoDB table: {dynamo_table_name}")
    dynamo_resource = boto3_resource('dynamodb', region_name="us-west-2")
    dynamo_table = dynamo_resource.Table(dynamo_table_name)
    return dynamo_table


def create_ddb_metadata_file_item(filename: str,
                                  algorithm_version: str,
                                  include_archive_time: bool = False,
                                  additional_metadata: dict = None):
    """Write metadata record to DynamoDB for a single file """
    ddb_metadata_item = {
        'PK': filename,
        'SK': "#",
        'algorithm-version': algorithm_version
    }
    if include_archive_time:
        add_archive_time_to_ddb_item(ddb_metadata_item)
    if additional_metadata:
        ddb_metadata_item.update(additional_metadata)
    logger.debug(f"Metadata record written to DynamoDB for {filename}")
    return ddb_metadata_item


def add_archive_time_to_ddb_item(ddb_item: dict):
    """Add archive time to DynamoDB item"""
    ddb_item.update({'archive-time': datetime.now(timezone.utc).isoformat()})
    return ddb_item


def create_ddb_metadata_applicable_date_item(filename: str,
                                             data_level: str,
                                             data_type: str,
                                             applicable_date: str,
                                             data_subtype: str = None,
                                             additional_metadata: dict = None):
    """Write metadata record to DynamoDB for a single file """
    sort_key = f"#{data_level}#{data_type}"
    if data_subtype:
        sort_key += f"#{data_subtype}"
    ddb_metadata_item = {
        'PK': filename,
        'SK': sort_key,
        'applicable-date': applicable_date
    }
    if additional_metadata:
        ddb_metadata_item.update(additional_metadata)
    logger.debug("Metadata applicable-date record written to DynamoDB")
    return ddb_metadata_item
